package cz.jzitnik.game;

import cz.jzitnik.game.items.Item;
import cz.jzitnik.game.items.ItemType;
import cz.jzitnik.game.sprites.Breaking;
import cz.jzitnik.tui.ScreenMovingCalculationProvider;
import cz.jzitnik.tui.ScreenRenderer;
import lombok.Getter;
import org.jline.terminal.Terminal;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.Random;

@Getter
public class Game {
    private List<Block>[][] world = new ArrayList[256][512];
    private Block player;
    private boolean mining = false;

    private List<Item> inventory;
    private Optional<Item> itemInHand = Optional.empty();

    public Game() {
        for (int i = 0; i < 256; i++) {
            for (int j = 0; j < 512; j++) {
                world[i][j] = new ArrayList<>();
            }
        }

        Block steveBlock = new Block("steve", SpriteLoader.SPRITES.STEVE);
        player = steveBlock;

        Random random = new Random();
        int baseHeight = 120; // Base ground level
        int[] terrainHeight = new int[512];

        // Generate terrain with gradual height variations
        terrainHeight[0] = baseHeight;
        for (int i = 1; i < 512; i++) {
            int heightChange = random.nextInt(3) - 1; // -1, 0, or +1
            terrainHeight[i] = Math.max(100, Math.min(140, terrainHeight[i - 1] + heightChange));
        }

        // Smooth terrain to avoid unnatural peaks and dips
        for (int i = 2; i < 510; i++) {
            terrainHeight[i] = (terrainHeight[i - 1] + terrainHeight[i] + terrainHeight[i + 1]) / 3;
        }

        // Generate world blocks based on terrain height
        for (int i = 0; i < 512; i++) {
            int hillHeight = terrainHeight[i];

            world[hillHeight][i].add(new Block("grass", SpriteLoader.SPRITES.GRASS));

            // Dirt layers
            for (int j = 1; j <= 4; j++) {
                if (hillHeight + j < 256) {
                    world[hillHeight + j][i].add(new Block("dirt", SpriteLoader.SPRITES.DIRT));
                    world[hillHeight + j][i].add(new Block("dirt", SpriteLoader.SPRITES.DIRT));
                }
            }

            // Stone layers below dirt
            for (int j = hillHeight + 5; j < 250; j++) {
                world[j][i].add(new Block("stone", SpriteLoader.SPRITES.STONE, 3));
            }

            // Bedrock at the bottom
            world[255][i].add(new Block("bedrock", SpriteLoader.SPRITES.BEDROCK));
        }

        // Fill empty spaces with air
        for (List<Block>[] lists : world) {
            for (List<Block> list : lists) {
                if (list.isEmpty()) {
                    list.add(new Block("air", SpriteLoader.SPRITES.AIR, true));
                }
            }
        }

        // Spawn player at a valid starting point
        world[terrainHeight[256] - 1][256].add(steveBlock);
    }

    public int calculateHardness(Block block) {
        int holdingDecrease = 0;
        if (itemInHand.isPresent() && block.getTool().isPresent() && itemInHand.get().getType().equals(block.getTool().get())) {
            holdingDecrease = itemInHand.get().getMiningDecrease();
        }

        int decrease = block.getHardness() - holdingDecrease;

        if (decrease < 0) {
            decrease = 0;
        }

        return decrease;
    }

    public int[] getPlayerCords() {
        for (int i = 0; i < world.length; i++) {
            for (int j = 0; j < world[i].length; j++) {
                for (Block block : world[i][j]) {
                    if (block.getBlockId().equals("steve")) {
                        return new int[]{j, i};
                    }
                }
            }
        }
        return null;
    }

    public void movePlayerRight(ScreenRenderer screenRenderer) {
        int[] cords = getPlayerCords();

        if (world[cords[1]][cords[0] + 1].stream().anyMatch(block -> !block.isGhost())) {
            return;
        }

        world[cords[1]][cords[0] + 1].add(player);
        world[cords[1]][cords[0]].remove(player);

        update(screenRenderer);
    }

    public void movePlayerLeft(ScreenRenderer screenRenderer) {
        int[] cords = getPlayerCords();

        if (world[cords[1]][cords[0] - 1].stream().anyMatch(block -> !block.isGhost())) {
            return;
        }

        world[cords[1]][cords[0] - 1].add(player);
        world[cords[1]][cords[0]].remove(player);

        update(screenRenderer);
    }

    public void movePlayerUp(ScreenRenderer screenRenderer) {
        int[] cords = getPlayerCords();

        if (world[cords[1] - 1][cords[0]].stream().anyMatch(block -> !block.isGhost()) || world[cords[1] + 1][cords[0]].stream().anyMatch(Block::isGhost)) {
            return;
        }

        world[cords[1] - 1][cords[0]].add(player);
        world[cords[1]][cords[0]].remove(player);

        new Thread(() -> {
            try {
                Thread.sleep(500);
            } catch (InterruptedException e) {
                throw new RuntimeException(e);
            }
            int[] cords2 = getPlayerCords();

            if (world[cords2[1] + 1][cords2[0]].stream().anyMatch(Block::isGhost)) {
                world[cords2[1] + 1][cords2[0]].add(player);
                world[cords2[1]][cords2[0]].remove(player);

                screenRenderer.render(world);
            }
        }).start();
    }

    public void mine(ScreenRenderer screenRenderer, int x, int y) {
        if (mining) {
            return;
        }

        Block breakingBlock = new Block("breaking", SpriteLoader.SPRITES.BREAKING);
        world[y][x].add(breakingBlock);
        screenRenderer.render(world);

        int hardness = calculateHardness(world[y][x].stream().filter(block -> !block.isGhost()).toList().get(0));

        this.mining = true;

        new Thread(() -> {
            try {
                Thread.sleep(hardness * 166L);
            } catch (InterruptedException e) {
                throw new RuntimeException(e);
            }
            breakingBlock.setSpriteState(Breaking.BreakingState.SECOND);
            screenRenderer.render(world);

            try {
                Thread.sleep(hardness * 166L);
            } catch (InterruptedException e) {
                throw new RuntimeException(e);
            }
            breakingBlock.setSpriteState(Breaking.BreakingState.THIRD);
            screenRenderer.render(world);

            try {
                Thread.sleep(hardness * 166L);
            } catch (InterruptedException e) {
                throw new RuntimeException(e);
            }

            mining = false;

            world[y][x].clear();
            world[y][x].add(new Block("air", SpriteLoader.SPRITES.AIR, true));
            screenRenderer.render(world);

            update(screenRenderer);
        }).start();
    }

    public boolean isMineable(int x, int y, Terminal terminal) {
        List<Block> blocks = world[y][x];

        int[] cords = getPlayerCords();

        int playerX = cords[0];
        int playerY = cords[1];

        int distanceX = Math.abs(playerX - x);
        int distanceY = Math.abs(playerY - y);

        int[] data = ScreenMovingCalculationProvider.calculate(playerX, playerY, terminal.getHeight(), terminal.getWidth(), world[0].length, world.length);

        int startX = data[0];
        int endX = data[1];
        int startY = data[2];
        int endY = data[3];

        return
                y >= startY && y < endY && x >= startX && x < endX &&
                !blocks.stream().allMatch(block -> block.getBlockId().equals("air"))
                && distanceX <= 5 && distanceY <= 5
                && !(playerX == x && playerY == y);
    }

    public void update(ScreenRenderer screenRenderer) {
        while (true) {
            try {
                Thread.sleep(200);
            } catch (InterruptedException e) {
                throw new RuntimeException(e);
            }
            int[] cords2 = getPlayerCords();

            if (world[cords2[1] + 1][cords2[0]].stream().allMatch(Block::isGhost)) {
                world[cords2[1] + 1][cords2[0]].add(player);
                world[cords2[1]][cords2[0]].remove(player);

                screenRenderer.render(world);
            } else {
                break;
            }
        }
    }
}